/* ====================================================================
||
|| vma - A utility to list and extract contents of VMARC archive files
||
||
|| ../vma - List/Extract files from VMARC archives
||
|| Usage: ../vma [options] input
||
|| Options:
||   -a        convert extracted files to ASCII (implies -t)
||   -c        convert names to lowercase
||   -h        display usage summary
||   -l        list files (default)
||   -m fm     replace filemode...0=remove
||   -q        do not list files
||   -s        safe, slow, subfile searching
||   -u f,t    specifies (f)rom and (t) UCM filenames
||   -x        extract files
||   -v        verbose listing
||   -V        display version
||
|| input:
||   name of the VMARC archive
||
|| fn, ft, fm:
||   filter on file name, type, and/or mode
||   (case is significant)
||
|| To compile with GCC:
||
||   see Makefile
||
|| To compile on MVS:
||
||   see Makefile.jcl
||
|| To compile on VM:
||
||   (anyone want to supply these?)
||
==================================================================== */

#define _ALL_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <errno.h>
#include <time.h>
#include <utime.h>
#include <unistd.h>

#include "version.h"
#include "vma.h"

/* --------------------------------------------------------------------
|| Silly getopt stuff
*/

#if !defined(__MVS__)
#include <getopt.h>
#else
#define optind *( __opindf() )
#define optarg *( __opargf() )
#endif

/* --------------------------------------------------------------------
|| General stuff
*/
#if !defined( NULL )
#define NULL 0
#endif

#if !defined( TRUE )
#define TRUE 1
#endif

#if !defined( FALSE )
#define FALSE 0
#endif

/* --------------------------------------------------------------------
|| (naive) EBCDIC -> ASCII stuff
*/

#if defined( __MVS__ )
#define OPT_A
#else
#define OPT_A "a"
#endif

/* --------------------------------------------------------------------
|| Process control stuff
*/
static char f_debug   = FALSE;              /* turn on debugging     */
static char f_case    = FALSE;              /* names > lowercase     */
static char f_list    = TRUE;               /* list subfile info     */
static char f_safe    = FALSE;              /* safe subfile search   */
static char f_trans   = FALSE;              /* translate subfile     */
static char f_verbose = FALSE;              /* enable verbose output */
static char f_version = FALSE;              /* display version       */
static char f_extract = FALSE;              /* extract subfiles      */
static int  xmode     = VMAX_BINARY;        /* extraction mode       */
static char *s_mode   = NULL;               /* convert mode to...    */
static char *s_fn     = "*";                /* file name filter      */
static char *s_ft     = "*";                /* file type filter      */
static char *s_fm     = "*";                /* file mode filter      */
static char *s_filter;                      /* filter string         */
static char *s_fucm   = NULL;               /* from UCM charmap      */
static char *s_tucm   = NULL;               /* to UCM charmap        */
static int i_flen;                          /* len of filter         */
static char *s_name;                        /* output name           */
static int i_nlen;                          /* len of name           */

#if defined( _WIN32 )
/* --------------------------------------------------------------------
|| Prevent MinGW automatic command line globbing
*/
int _CRT_glob = 0;
#endif

/*
|| Since MinGW doesn't have an fnmatch() function, I'm obliged to
|| include a little globber.  At least it will be consistent across
|| platforms.
||
|| This particular version is in the public domain and was found at:
||     http://www.cs.yorku.ca/~oz
||
|| robust glob pattern matcher
|| ozan s. yigit/dec 1994
|| public domain
||
|| glob patterns:
||      *       matches zero or more characters
||      ?       matches any single character
||      [set]   matches any character in the set
||      [^set]  matches any character NOT in the set
||              where a set is a group of characters or ranges. a range
||              is written as two characters seperated with a hyphen:
||              a-z denotes all characters between a to z inclusive.
||      [-set]  set matches a literal hypen and any character in the set
||      []set]  matches a literal close bracket and any character in the
||              set
||
||      char    matches itself except where char is '*' or '?' or '['
||      \char   matches char, including any pattern character
||
|| examples:
||      a*c             ac abc abbc ...
||      a?c             acc abc aXc ...
||      a[a-z]c         aac abc acc ...
||      a[-a-z]c        a-c aac abc ...
||
|| $Log: glob.c,v $
|| Revision 1.3  1995/09/14  23:24:23  oz
|| removed boring test/main code.
||
|| Revision 1.2  94/12/11  10:38:15  oz
|| cset code fixed. it is now robust and interprets all
|| variations of cset [i think] correctly, including [z-a] etc.
||
|| Revision 1.1  94/12/08  12:45:23  oz
|| Initial revision
*/
#ifndef NEGATE
#define NEGATE  '^'                     /* std cset negation char */
#endif

static int
amatch( char *str, char *p )
{
    int negate;
    int match;
    int c;

    while( *p )
    {
        if( !*str && *p != '*' )
        {
            return FALSE;
        }

        switch( c = *p++ )
        {
            case '*':
                while( *p == '*' )
                {
                    p++;
                }

                if( !*p )
                {
                    return TRUE;
                }

                if( *p != '?' && *p != '[' && *p != '\\' )
                {
                    while( *str && *p != *str )
                    {
                        str++;
                    }
                }

                while( *str )
                {
                    if( amatch( str, p ) )
                    {
                        return TRUE;
                    }
                    str++;
                }

            return FALSE;

            case '?':
                if( *str )
                {
                    break;
                }

            return FALSE;

            /*
            || set specification is inclusive, that is [a-z] is a, z
            || and everything in between. this means [z-a] may be
            || interpreted as a set that contains z, a and nothing in
            || between.
            */
            case '[':
                if( *p != NEGATE )
                {
                    negate = FALSE;
                }
                else
                {
                    negate = TRUE;
                    p++;
                }

                match = FALSE;

                while( !match && ( c = *p++ ) )
                {
                    if( !*p )
                    {
                        return FALSE;
                    }

                    if( *p == '-' )         /* c-c */
                    {
                        if( !*++p )
                        {
                            return FALSE;
                        }

                        if( *p != ']' )
                        {
                            if( *str == c || *str == *p ||
                                (*str > c && *str < *p ) )
                            {
                                   match = TRUE;
                            }
                        }
                        else                /* c-] */
                        {
                            if( *str >= c )
                            {
                                match = TRUE;
                            }
                            break;
                        }
                    }
                    else                    /* cc or c] */
                    {
                        if( c == *str )
                        {
                            match = TRUE;
                        }

                        if( *p != ']' )
                        {
                            if( *p == *str )
                            {
                                match = TRUE;
                            }
                        }
                        else
                        {
                            break;
                        }
                    }
                }

                if( negate == match )
                {
                    return FALSE;
                }

                /*
                || if there is a match, skip past the cset and
                || continue on
                */
                while( *p && *p != ']' )
                {
                    p++;
                }

                if( !*p++ )      /* oops! */
                {
                    return FALSE;
                }
            break;

            case '\\':
                if( *p )
                {
                    c = *p++;
                }

            /* intentional fallthrough */

            default:
                if( c != *str )
                {
                    return FALSE;
                }
            break;
        }
        str++;
    }

    return !*str;
}

static void
make_name( SUBFILE *sf )
{
    char *mode;
    int dot;

    mode = (char *) sf->fm;
    dot = '.';
    if( s_mode )
    {
        mode = s_mode;
        if( strcmp( s_mode, "0" ) == 0 )
        {
            dot = '\0';
        }
    }

    sprintf( s_name,
             "%s.%s%c%s",
             sf->fn,
             sf->ft,
             dot,
             mode );

    if( f_case )
    {
        mode = s_name;
        while( *mode )
        {
            *mode = tolower( *mode );
            mode++;
        }
    }

    return;
}

static void
list_file( SUBFILE *sf )
{
    static int needhead = TRUE;

    /*
    || Only print the header once
    */
    if( needhead )
    {
        printf( "Fn       " );
        printf( "Ft       " );
        printf( "Fm " );
        if( f_verbose )
        {
            printf( "V.R " );
            printf( "Meth " );
        }
        printf( "Date       " );
        printf( "Time     " );
        printf( "R " );
        printf( "Lrecl " );
        printf( " Bytes In " );
        printf( "Bytes Out\n" );

        needhead = FALSE;
    }

    /*
    || Print the first few common fields
    */
    printf( "%-8.8s ",
            sf->fn );

    printf( "%-8.8s ",
            sf->ft );

    printf( "%-2.2s ",
            sf->fm );

    /*
    || Print some extra info if being verbose
    */
    if( f_verbose )
    {
        printf( "%1d.%1d ",
                sf->ver,
                sf->rel );

        printf( "%-4.4s ",
                sf->meth );
    }

    /*
    || Print the remaining fields
    */
    printf( "%04d/%02d/%02d %02d:%02d:%02d ",
            sf->year,
            sf->month,
            sf->day,
            sf->hour,
            sf->minute,
            sf->second );

    printf( "%c ",
            sf->recfm );

    printf( "%5d ",
            sf->lrecl );

    printf( "%9d %9d\n",
            sf->ibytes,
            sf->obytes );

    return;
}

static void
usage( void )
{
    printf( "vma - List/Extract files from VMARC archives\n\n" );
    printf( "Usage: vma [options] input [fn [ft [fm ]]]\n\n" );
    printf( "Options:\n" );

#if !defined( __MVS__ )
    printf( "  -a        convert extracted files to ASCII\n" );
#endif

    printf( "  -c        convert names to lowercase\n" );
    printf( "  -h        display usage summary\n" );
    printf( "  -l        list files (default)\n" );
    printf( "  -m fm     replace filemode...0=remove\n" );
    printf( "  -q        do not list files\n" );
    printf( "  -s        safe, slow, subfile searching\n" );
    printf( "  -u f,t    (f)rom and (t) UCM filenames\n" );
    printf( "  -v        verbose listing\n" );
    printf( "  -x        extract files\n" );
    printf( "  -V        display version\n\n" );
    printf( "input:\n"
            "  name of the VMARC archive\n\n" );
    printf( "fn, ft, fm:\n"
            "  filter on file name, type, and/or mode\n" );
    printf( "  (case is significant)\n\n" );
    printf( "f,t:\n" );
    printf( "  paths to translateion tables (see README)\n" );

    exit( 1 );
}

int
main( int argc, char *argv[] )
{
    /*          Fn  .   Ft  .   Fm  0 */
    char fname[ 8 + 1 + 8 + 1 + 2 + 1 ];
    void *vma;
    char *p;
    int rc;
    int cnt;
    int sfcount = 0;
    int sfproc = 0;

    /*
    || Process command flags
    */
    while( ( rc = getopt( argc, argv, OPT_A "chlm:qstu:vxV" ) ) != -1 )
    {
        switch( rc )
        {
#if !defined( __MVS__ )
            case 'a':
                xmode = VMAX_TEXT;
            break;
#endif
            case 'c':
                f_case = TRUE;
            break;

            case 'l':
                f_list = TRUE;
            break;

            case 'm':
                s_mode = optarg;
            break;

            case 't':
                xmode = VMAX_TRANS;
            break;

            case 'u':
                s_fucm = optarg;
                s_tucm = strchr( optarg, ',' );
                if( !s_tucm )
                {
                    usage();
                }
                (*s_tucm++) = '\0';
            break;

            case 'q':
                f_list = FALSE;
            break;

            case 's':
                f_safe = TRUE;
            break;

            case 'V':
                f_version = TRUE;
            break;

            case 'v':
                f_verbose = TRUE;
            break;

            case 'x':
                f_extract = TRUE;
            break;

            case 'h':
            default:
                usage();
            break;
        }
    }

    if( f_version )
    {
        printf( "%s\nVersion " VERSION
                " - compiled on " __DATE__
                " at " __TIME__ "\n",
                argv[ 0 ] );

        exit( 0 );
    }

    /*
    || Make sure we have the right number of arguments
    */
    cnt = argc - optind;
    if( cnt < 1 || cnt > 4 )
    {
        usage();
    }

    /*
    || User specified filter?
    */
    if( cnt > 1 )
    {
        /*
        || First optional argument becomes the file name filter
        */
        s_fn = argv[ optind + 1 ];

        /*
        || Second optional argument becomes the file type filter
        */
        if( cnt > 2 )
        {
            s_ft = argv[ optind + 2 ];

            /*
            || Third optional argument becomes the file mode filter
            */
            if( cnt > 3 )
            {
                s_fm = argv[ optind + 3 ];
            }
        }
    }

    /*
    || Calculate filter length
    */
    i_flen = strlen( s_fn ) + 1 +
             strlen( s_ft ) + 1 +
             strlen( s_fm ) + 1;

    /*
    || Allocate it
    */
    s_filter = malloc( i_flen );
    if( s_filter == NULL )
    {
        printf( "no mem\n" );
        exit( 1 );
    }

    /*
    || Build the filter string
    */
    sprintf( s_filter,
             "%s.%s.%s",
             s_fn,
             s_ft,
             s_fm );

    /*
    || Just in case the user forgot what file was being worked on
    */
    if( f_verbose )
    {
        printf( "Processing: %s\n\n", argv[ optind ] );
    }

    /*
    || Display non-default filter
    */
    if( strcmp( s_filter, "*.*.*" ) != 0 )
    {
        printf("Using filter: '%s'\n\n", s_filter );
    }

    /*
    || Calculate output name length
    */
    i_nlen = 8 + 1 +
             8 + 1 +
             ( s_mode ? strlen( s_mode ) : 8 ) + 1;

    /*
    || Allocate it
    */
    s_name = malloc( i_nlen );
    if( s_name == NULL )
    {
        printf( "no mem\n" );
        exit( 1 );
    }

    /*
    || Open the VMARC file
    */
    rc = vma_open( argv[ optind ], f_safe, &vma );
    while( rc == VMAE_NOERR )
    {
        SUBFILE *sf;

        if( s_fucm )
        {
            rc = vma_setconv( vma, s_fucm, s_tucm );
            if( rc != VMAE_NOERR )
            {
                printf( "Failed to set conversion: %d\n", rc );
                break;
            }
        }

        for( rc = vma_first( vma, &sf );
             rc == VMAE_NOERR;
             rc = vma_next( vma, &sf ) )
        {
            /*
            || Track total subfile count
            */
            sfcount++;

            /*
            || Build a name for filtering
            */
            sprintf( fname,
                     "%s.%s.%s",
                     sf->fn,
                     sf->ft,
                     sf->fm );

            /*
            || Filter it
            */
            if( !amatch( fname, s_filter ) )
            {
                continue;
            }

            /*
            || Extract file
            */
            if( f_extract )
            {
                /*
                || Build the output name
                */
                make_name( sf );

                /*
                || And extract
                */
                rc = vma_extract( vma, s_name, xmode );
                if( rc != VMAE_NOERR )
                {
                    if( rc != VMAE_LRECL )
                    {
                        break;
                    }

                    printf( "Bypassing next file due "
                            "to LRECL limitations:\n" );
                }
            }

            /*
            || List it ... do after extraction to get byte counts
            */
            if( f_list )
            {
                list_file( sf );
            }

            /*
            || Track processed subfile count
            */
            sfproc++;

        }

        /*
        || This one really isn't an error
        */
        if( rc == VMAE_NOMORE )
        {
            rc = VMAE_NOERR;
        }

        /*
        || Always force exit from loop
        */
        break;
    }

    /*
    || Print a little summary
    */
    if( f_verbose )
    {
        printf( "\n%d subfiles",
                sfcount );

        if( sfcount != sfproc )
        {
            printf( ", %d bypassed due to filtering",
                    sfcount - sfproc );
        }

        printf( "\n" );
    }

    if( vma )
    {
        vma_close( vma );
    }

    return rc;
}
